#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>



MODULE_LICENSE ("GPL");
MODULE_DESCRIPTION ("Copia strings y los muestra con un arreglo doble");
MODULE_AUTHOR ("Eduardo Ruiz Duarte");


#define MAX_FRASE_SIZE 32
#define MAX_FRASES 32

static struct proc_dir_entry *proc_entry;

static char **cadena;		
static int cadena_index;	
static int siguiente_frase;
ssize_t
cadena_write (struct file *filp, const char __user * buff,
	      unsigned long len, void *data)
{
  if (len > MAX_FRASE_SIZE)
    {
      printk (KERN_INFO "cadena: Frase muy grande\n");
      return -ENOSPC;
    }
  if (cadena_index > MAX_FRASES)
    {
      printk (KERN_INFO "cadena: Ya se lleno el arreglo de cadena\n");
      return -ENOSPC;
    }
  printk (KERN_INFO "cadena: Alojando memoria\n");
  cadena[cadena_index] = vmalloc (len + 1);
  printk (KERN_INFO "cadena: Direccion de memoria de slot de frase nueva %p\n",
	  cadena[cadena_index]);
  if (!cadena)
    {
      printk (KERN_INFO "cadena: Ya no hay memoria o algo raro paso\n");
      return -ENOMEM;
    }
  memset (cadena[cadena_index], 0, len + 1);
  printk (KERN_INFO "cadena: Copiando de userspace\n");
  if (copy_from_user (cadena[cadena_index], buff, len))
    {
      return -EFAULT;
    }
  printk (KERN_INFO "cadena: Copiado de userspace\n");
  cadena_index++;
  return len;
}


int
cadena_read (char *page, char **start, off_t off,
	     int count, int *eof, void *data)
{
  int len;
  if (off > 0)
    {
      *eof = 1;
      return 0;
    }

  printk (KERN_INFO "cadena: s=%d,i=%d\n", siguiente_frase, cadena_index);
  if (cadena_index != 0)
    {
  siguiente_frase %= cadena_index;
      len = sprintf (page, "%s\n", cadena[siguiente_frase]);
      siguiente_frase++;
    }
  else {
    printk (KERN_INFO, "cadena: No hay cadena\n");
    return 0;
  }
  return len;
}

int
init_cadena_module (void)
{
  int ret = 0, i;
  cadena = (char *) vmalloc (sizeof (char *) * MAX_FRASES);
  printk (KERN_INFO "cadena: Direccion de mem de cadena %p\n", cadena);
  if (!cadena)
    {
      printk (KERN_INFO "cadena: no hay memoria para iniciar\n");
      ret = -ENOMEM;
    }
  else
    {
      proc_entry = create_proc_entry ("cadena", 0644, NULL);
      if (proc_entry == NULL)
	{
	  ret = -ENOMEM;
	  for (i = 0; i < cadena_index; i++)
	    vfree (cadena[i]);
	  vfree (cadena);

	  printk (KERN_INFO "cadena: No se pudo crear procentry\n");
	}
      else
	{

	  cadena_index = 0;
	  siguiente_frase = 0;
	  proc_entry->read_proc = cadena_read;
	  proc_entry->write_proc = cadena_write;
	  printk (KERN_INFO "cadena: modulo cargado\n");
	}
    }
  return ret;
}

void
cleanup_cadena_module (void)
{
  int i;
  for (i = 0; i < cadena_index; i++)
    vfree (cadena[i]);
  remove_proc_entry ("cadena", &proc_entry);
  vfree (cadena);
  printk (KERN_INFO "cadena: modulo descargado\n");
}

module_init (init_cadena_module);
module_exit (cleanup_cadena_module);
